{-
	InGraph - Ingress link optimizer

    Copyright (C) 2013  Nigel D. Stepp

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Nigel Stepp <stepp@atistar.net>
	http://www.atistar.net/~stepp/contactxfer/

    $Id: InGraph.hs 819 2013-03-16 00:47:09Z stepp $
-}

-- |Ingress Graph module.
-- The routines in this module support optimizing links between
-- portals in the game Ingress.
module InGraph where

import Data.List
import qualified Data.Set as Set
import Data.Maybe
import System.Random
import Data.Graph.Inductive

-- * Types

data Vector a = Vec (a,a) deriving (Eq, Show)
type Portal = (String,Vector Float)
type PortalGraph = Gr Portal ()

instance Functor Vector where
    fmap f (Vec (x,y)) = Vec (f x, f y)

-- | Allow math to be used with a 'Vector'
instance Num a => Num (Vector a) where
    (+) (Vec (a1,a2)) (Vec (b1,b2)) = Vec (a1+b1, a2+b2)
    (-) (Vec (a1,a2)) (Vec (b1,b2)) = Vec (a1-b1, a2-b2)
    (*) (Vec (a1,a2)) (Vec (b1,b2)) = Vec (a1*b1, a2*b2)
    negate (Vec (a1,a2)) = Vec (-a1,-a2)
    abs (Vec (a1,a2)) = Vec (abs a1, abs a2)
    signum (Vec (a1,a2)) = Vec (signum a1, signum a2)
    fromInteger x = Vec (fromInteger x, fromInteger x)

-- * Functions

-- ** Graph handling routines

emptyGraph = empty::PortalGraph

-- |Add a list of portals to a portal graph
addPortals :: PortalGraph -> [Portal] -> PortalGraph
addPortals g [] = g
addPortals g (p:ps) = addPortals (insNode (newNode g, p) g) ps

-- |Determine the id of the next node to add
newNode :: PortalGraph -> Node
newNode g = (newNodes 1 g)!!0

-- |Remove all of the edges from a graph
stripEdges :: PortalGraph -> PortalGraph
stripEdges g =
    delEdges (edges g) g

-- ** Mathematical Utilities

-- |Arithmetic mean
mean x = (sum x) / (fromIntegral $ length x)

-- |Standard deviation
std x = (foldl (\b c -> b + c^2) 0 (map (\a -> a - (mean x)) x)) / (fromIntegral $ length x - 1)

-- |Two dimensional cross product, used for link crossing
cross2D :: Num a => Vector a -> Vector a -> a
cross2D (Vec (v1,v2)) (Vec (w1,w2)) = v1*w2 - v2*w1


-- ** Graph queries

-- |Determine whether or not links between two pairs of portals
-- cross
linksCross :: (Portal,Portal) -> (Portal,Portal) -> Bool
linksCross link1 link2 =
    let p1 = snd (fst link1)
        p2 = snd (snd link1)
        p3 = snd (fst link2)
        p4 = snd (snd link2)
        v = p2 - p1
        w = p4 - p3
        denom = cross2D v w
    in if denom /= 0
       then let crossPointv = (cross2D (p3 - p1) w) / denom
                crossPointw = (cross2D (p3 - p1) v) / denom
            in (crossPointw > 0.0 && crossPointw < 1.0) && (crossPointv > 0.0 && crossPointv < 1.0)
       else False

-- |Count the number of fields adjacent to a portal.
-- Returns a list of endpoint pairs for each triangle.
-- For instance:
--
-- >>> adjFields g 1
-- [(2,3),(3,4)]
-- 
-- Specifies that node 1 makes triangles (1,2,3) and (1,3,4).
adjFields :: PortalGraph   -- ^ Graph to analyze
          -> Node          -- ^ Find fields with this node as a vertex
          -> [(Node,Node)] -- ^ A list containing pairs of vertices that
                           -- make a trianlge with the given reference vertex
adjFields g p = let links = suc (undir g) p
                    endpoints = map (intersect links) (map (suc g) links)
                    endpairs = map (\a -> zip (repeat (fst a)) (snd a) ) (zip links endpoints)
                in concat endpairs

-- |Calculate the area covered by a field
fieldArea :: Portal -> Portal -> Portal -> Float
fieldArea portal1 portal2 portal3 =
    let p1 = snd portal1
        p2 = snd portal2
        p3 = snd portal3
        a = norm (p2-p1)
        b = norm (p3-p2)
        c = norm (p1-p3)
        s = (a+b+c)/2
    in sqrt (s*(s-a)*(s-b)*(s-c))
    where norm (Vec (x,y)) = sqrt (x^2+y^2)

-- |Calculate the area covered by a field,
-- using graph node numbers
fieldAreaG :: PortalGraph -> Node -> Node -> Node -> Float
fieldAreaG g p1 p2 p3 =
    let portal1 = fromJust $ lab g p1
        portal2 = fromJust $ lab g p2
        portal3 = fromJust $ lab g p3
    in fieldArea portal1 portal2 portal3

-- |Determine whether a potential link would cross
-- an existing link
makesCross :: PortalGraph -> (Node,Node) -> Bool
makesCross g (from,to) =
    let fromNode = fromJust $ lab g from
        toNode = fromJust $ lab g to
        crosses = map (\(a,b)->linksCross (fromNode,toNode) (fromJust $ lab g a, fromJust $ lab g b)) (edges g)
    in or crosses


-- ** Optimization

-- |Determine the energy of a graph
hamiltonian :: Float -> PortalGraph -> Float
hamiltonian gain g = let energies = map (portalEnergy gain g) (nodes g)
                in (sum energies)

-- |Calculate the energy of a single portal.
-- Energy is defined to be the ratio of AP
-- due to destruction over AP due to creation.
-- This is then scaled by the number of fields.
portalEnergy :: Float       -- ^ Importance of making fields, 0 to 1
             -> PortalGraph -- ^ The current graph
             -> Node        -- ^ Calculate the energy of this node
             -> Float       -- ^ Returns portal energy
portalEnergy fieldGain g p =
    let links = deg g p
        fields = adjFields g p
        numFields = length fields
        --area = foldl (\a f -> a + (fieldAreaG g p (fst f) (snd f))) 0 fields
        enemyAP = fromIntegral (187*links + 750*numFields)
        friendlyAP = fromIntegral (313*links + 1250*numFields)
    in if links < 1
          then 1e6
          else enemyAP / friendlyAP / (1-fieldGain+fieldGain*(tanh $ fromIntegral (1+numFields)))

-- |Implement a simple metropolis algorithm for minimizing
-- the hamiltonian of a portal graph
metropolis :: StdGen      -- ^ Random number generator
           -> PortalGraph -- ^ Graph to optimize
           -> Float       -- ^ Importance of making fields, 0 to 1
           -> Int         -- ^ Maximum number of iterations
           -> PortalGraph -- ^ Returns the optimized graph
metropolis randgen g gain 0 = g
metropolis randgen g gain n =
    let (temp,newRandgen) = perturbGraph randgen g
    in if (hamiltonian gain temp) < (hamiltonian gain g)
          then metropolis newRandgen temp gain (n-1)
          else metropolis newRandgen g gain (n-1)

-- |Create a random modification to a graph.
-- Pick a link to toggle on or off.
-- Continue toggling links until one of the toggles adds a link.
-- This is because the hamiltonian will nearly always judge removal of a link as a higher energy state.
perturbGraph :: StdGen -> PortalGraph -> (PortalGraph,StdGen)
perturbGraph randgen g = 
    -- Select two random nodes for a link
    let (randLinkFrom,randgen2) =  randomR (1, noNodes g) randgen
        (randLinkTo,randgen3) =  randomR (1, noNodes g) randgen2
    --- If it's a self-link, try again
    in if randLinkFrom == randLinkTo
       then if (length $ nodes g) < 2
            then (g,randgen3)
            else perturbGraph randgen3 g
       else 
          --Toggle links until a toggle adds a link
          if (randLinkFrom,randLinkTo) `elem` (edges g)
             then perturbGraph randgen3 (delEdge (randLinkFrom,randLinkTo) g)
             else if (randLinkTo,randLinkFrom) `elem` (edges g)
                  then perturbGraph randgen3 (delEdge (randLinkTo,randLinkFrom) g)
                  else if makesCross g (randLinkFrom,randLinkTo)
                       then perturbGraph randgen3 g
                       else (insEdge (randLinkFrom,randLinkTo,()) g, randgen3)
 
-- |Optimize the links among a collection of nodes.
graphOptimize :: Float       -- ^ fieldGain: Importance of making fields, from 0 to 1
              -> Int         -- ^ maxIter: Maximum optimzation iterations
              -> PortalGraph -- ^ portalGraph: Graph to optimize
              -> PortalGraph -- ^ Returns optimized graph
graphOptimize gain maxiter portalGraph =
    metropolis (mkStdGen 1) portalGraph gain maxiter


-- ** Graph measures

-- |Count the total number of fields created by links in a graph
countFields :: PortalGraph -> Int
countFields g =
    let allAdjFields = map (adjFields g) (nodes g)
        fieldVerts = zipWith (\a b -> map (\c-> [a, (fst c), (snd c)]) b)  (nodes g) allAdjFields
        uniqueFields = nub $ map Set.fromList $ concat fieldVerts
    in length uniqueFields

-- |Rank portals by importance. See "rankPortalNodes".
rankPortals :: PortalGraph -> [Portal]
rankPortals g = map (fromJust . lab g) (rankPortalNodes g)

-- |Rank portal nodes according to how important they are.
-- Importance is ranked by AP due to destruction.
rankPortalNodes :: PortalGraph -> [Node]
rankPortalNodes g =
    let links = map (deg g) (nodes g)
        fields = map (length . adjFields g) (nodes g)
        aps = zipWith (\x y -> 187*x + 750*y) links fields
        apsIdx = zip aps (nodes g)
    in map snd $ reverse $ sort apsIdx

    

-- |Calculate some descriptive numbers summarizing a graph.
-- Return a tuple of (Enemy AP, Friendly AP, Number of Fields)
graphStats :: PortalGraph     -- ^ The graph to analyze
           -> (Int, Int, Int) -- ^ Returns enemy AP, friendly AP,
                              --and the number of fields as a tuple
graphStats g =
    let links = sum $ map (outdeg g) (nodes g)
        numFields = countFields g
        enemyAP = 187*links + 750*numFields
        friendlyAP = 313*links + 1250*numFields
    in (enemyAP,friendlyAP,numFields)
     

-- * Misc


-- Sample portal list, used for testing
portals = [
--    ("Montana Ave Library", Vec (17.1,10.0)),
--    ("Douglas park", Vec (25.0,6.1)),
--    ("Mural (new balance)", Vec (26.1,6.0)),
--    ("Artistic house", Vec (26.0,7.0)),
--    ("St. Monica (church)", Vec (7.1,7.1)),
--    ("Post Office (26th St)", Vec (26.0,14.1)),
--    ("Museum Systems", Vec (27.0,14.0))
    ("Montana Ave Library", Vec (430,232)),
    ("Douglas park", Vec (588,221)),
    ("Mural (new balance)", Vec (605,213)),
    ("Artistic house", Vec (581,191)),
    ("St. Monica (church)", Vec (372,395)),
    ("Post Office (26th St)", Vec (447,51)),
    ("Museum Systems", Vec (461,41)),
    ("Palisades Branch Library", Vec (46,102)),
    ("Fire Station 69", Vec (77,86)),
    ("Post Office (Palisades)", Vec (40,72))
    ]

-- |Sample portal graph, for testing
northSaMo = addPortals emptyGraph portals


